"use strict";
Object.defineProperty(exports, "__esModule", {
    value: true
});
function _export(target, all) {
    for(var name in all)Object.defineProperty(target, name, {
        enumerable: true,
        get: all[name]
    });
}
_export(exports, {
    flattenBranches: ()=>flattenBranches,
    serializeDefinitionIfAllowed: ()=>serializeDefinitionIfAllowed,
    serializeCase: ()=>serializeCase
});
const _unionTs = require("../parse/ast/union.js");
const _domainsTs = require("../utils/domains.js");
const _errorsTs = require("../utils/errors.js");
const _genericsTs = require("../utils/generics.js");
const _objectKindsTs = require("../utils/objectKinds.js");
const _pathsTs = require("../utils/paths.js");
const _serializeTs = require("../utils/serialize.js");
const _branchTs = require("./branch.js");
const _composeTs = require("./compose.js");
const _propsTs = require("./rules/props.js");
const flattenBranches = (branches, ctx)=>{
    const discriminants = calculateDiscriminants(branches, ctx);
    const indices = branches.map((_, i)=>i);
    return discriminate(branches, indices, discriminants, ctx);
};
const discriminate = (originalBranches, remainingIndices, discriminants, ctx)=>{
    if (remainingIndices.length === 1) {
        return (0, _branchTs.flattenBranch)(originalBranches[remainingIndices[0]], ctx);
    }
    const bestDiscriminant = findBestDiscriminant(remainingIndices, discriminants);
    if (!bestDiscriminant) {
        return [
            [
                "branches",
                remainingIndices.map((i)=>branchIncludesMorph(originalBranches[i], ctx.type.scope) ? (0, _errorsTs.throwParseError)((0, _unionTs.writeUndiscriminatableMorphUnionMessage)(`${ctx.path}`)) : (0, _branchTs.flattenBranch)(originalBranches[i], ctx))
            ]
        ];
    }
    const cases = {};
    for(const caseKey in bestDiscriminant.indexCases){
        const nextIndices = bestDiscriminant.indexCases[caseKey];
        cases[caseKey] = discriminate(originalBranches, nextIndices, discriminants, ctx);
        if (caseKey !== "default") {
            pruneDiscriminant(cases[caseKey], bestDiscriminant.path, bestDiscriminant, ctx);
        }
    }
    return [
        [
            "switch",
            {
                path: bestDiscriminant.path,
                kind: bestDiscriminant.kind,
                cases
            }
        ]
    ];
};
const pruneDiscriminant = (entries, segments, discriminant, ctx)=>{
    for(let i = 0; i < entries.length; i++){
        const [k, v] = entries[i];
        if (!segments.length) {
            if (discriminant.kind === "domain") {
                if (k === "domain" || k === "domains") {
                    entries.splice(i, 1);
                    return;
                } else if (k === "class" || k === "value") {
                    // these keys imply a domain, but also add additional
                    // information, so can't be pruned
                    return;
                }
            } else if (discriminant.kind === k) {
                entries.splice(i, 1);
                return;
            }
        } else if ((k === "requiredProp" || k === "prerequisiteProp" || k === "optionalProp") && v[0] === segments[0]) {
            if (typeof v[1] === "string") {
                if (discriminant.kind !== "domain") {
                    return throwPruneFailure(discriminant);
                }
                entries.splice(i, 1);
                return;
            }
            pruneDiscriminant(v[1], segments.slice(1), discriminant, ctx);
            if (v[1].length === 0) {
                entries.splice(i, 1);
            }
            return;
        }
        // check for branch keys, which must be traversed even if there are no
        // segments left
        if (k === "domains") {
            if ((0, _genericsTs.keyCount)(v) !== 1 || !v.object) {
                return throwPruneFailure(discriminant);
            }
            pruneDiscriminant(v.object, segments, discriminant, ctx);
            return;
        } else if (k === "switch") {
            for(const caseKey in v.cases){
                pruneDiscriminant(v.cases[caseKey], segments, discriminant, ctx);
            }
            return;
        } else if (k === "branches") {
            for (const branch of v){
                pruneDiscriminant(branch, segments, discriminant, ctx);
            }
            return;
        }
    }
    return throwPruneFailure(discriminant);
};
const throwPruneFailure = (discriminant)=>(0, _errorsTs.throwInternalError)(`Unexpectedly failed to discriminate ${discriminant.kind} at path '${discriminant.path}'`);
const discriminantKinds = {
    domain: true,
    class: true,
    value: true
};
const calculateDiscriminants = (branches, ctx)=>{
    const discriminants = {
        disjointsByPair: {},
        casesByDisjoint: {}
    };
    for(let lIndex = 0; lIndex < branches.length - 1; lIndex++){
        for(let rIndex = lIndex + 1; rIndex < branches.length; rIndex++){
            const pairKey = `${lIndex}/${rIndex}`;
            const pairDisjoints = [];
            discriminants.disjointsByPair[pairKey] = pairDisjoints;
            const intersectionState = new _composeTs.IntersectionState(ctx.type, "|");
            (0, _branchTs.branchIntersection)(branches[lIndex], branches[rIndex], intersectionState);
            for(const path in intersectionState.disjoints){
                if (path.includes(_propsTs.mappedKeys.index)) {
                    continue;
                }
                const { l , r , kind  } = intersectionState.disjoints[path];
                if (!(0, _genericsTs.isKeyOf)(kind, discriminantKinds)) {
                    continue;
                }
                const lSerialized = serializeDefinitionIfAllowed(kind, l);
                const rSerialized = serializeDefinitionIfAllowed(kind, r);
                if (lSerialized === undefined || rSerialized === undefined) {
                    continue;
                }
                const qualifiedDisjoint = path === "/" ? kind : `${path}/${kind}`;
                pairDisjoints.push(qualifiedDisjoint);
                if (!discriminants.casesByDisjoint[qualifiedDisjoint]) {
                    discriminants.casesByDisjoint[qualifiedDisjoint] = {
                        [lSerialized]: [
                            lIndex
                        ],
                        [rSerialized]: [
                            rIndex
                        ]
                    };
                    continue;
                }
                const cases = discriminants.casesByDisjoint[qualifiedDisjoint];
                const existingLBranch = cases[lSerialized];
                if (!existingLBranch) {
                    cases[lSerialized] = [
                        lIndex
                    ];
                } else if (!existingLBranch.includes(lIndex)) {
                    existingLBranch.push(lIndex);
                }
                const existingRBranch = cases[rSerialized];
                if (!existingRBranch) {
                    cases[rSerialized] = [
                        rIndex
                    ];
                } else if (!existingRBranch.includes(rIndex)) {
                    existingRBranch.push(rIndex);
                }
            }
        }
    }
    return discriminants;
};
const parseQualifiedDisjoint = (qualifiedDisjoint)=>{
    const path = _pathsTs.Path.fromString(qualifiedDisjoint);
    return [
        path,
        path.pop()
    ];
};
const findBestDiscriminant = (remainingIndices, discriminants)=>{
    let bestDiscriminant;
    for(let i = 0; i < remainingIndices.length - 1; i++){
        const lIndex = remainingIndices[i];
        for(let j = i + 1; j < remainingIndices.length; j++){
            const rIndex = remainingIndices[j];
            const candidates = discriminants.disjointsByPair[`${lIndex}/${rIndex}`];
            for (const qualifiedDisjoint of candidates){
                const indexCases = discriminants.casesByDisjoint[qualifiedDisjoint];
                const filteredCases = {};
                const defaultCases = [
                    ...remainingIndices
                ];
                let score = 0;
                for(const caseKey in indexCases){
                    const filteredIndices = indexCases[caseKey].filter((i)=>{
                        const remainingIndex = remainingIndices.indexOf(i);
                        if (remainingIndex !== -1) {
                            delete defaultCases[remainingIndex];
                            return true;
                        }
                    });
                    if (filteredIndices.length === 0) {
                        continue;
                    }
                    filteredCases[caseKey] = filteredIndices;
                    score++;
                }
                const defaultCaseKeys = (0, _genericsTs.objectKeysOf)(defaultCases);
                if (defaultCaseKeys.length) {
                    filteredCases["default"] = defaultCaseKeys.map((k)=>parseInt(k));
                }
                if (!bestDiscriminant || score > bestDiscriminant.score) {
                    const [path, kind] = parseQualifiedDisjoint(qualifiedDisjoint);
                    bestDiscriminant = {
                        path,
                        kind,
                        indexCases: filteredCases,
                        score
                    };
                    if (score === remainingIndices.length) {
                        // if we find a candidate that discriminates all branches, return early
                        return bestDiscriminant;
                    }
                }
            }
        }
    }
    return bestDiscriminant;
};
const serializeDefinitionIfAllowed = (kind, definition)=>{
    switch(kind){
        case "value":
            return serializeIfPrimitive(definition);
        case "domain":
            return definition;
        case "class":
            return typeof definition === "string" ? definition : undefined;
        default:
            return;
    }
};
const serializeIfPrimitive = (data)=>{
    const domain = (0, _domainsTs.domainOf)(data);
    return domain === "object" || domain === "symbol" ? undefined : (0, _serializeTs.serializePrimitive)(data);
};
const serializeData = {
    value: (data)=>serializeIfPrimitive(data) ?? "default",
    class: (data)=>(0, _objectKindsTs.objectKindOf)(data) ?? "default",
    domain: _domainsTs.domainOf
};
const serializeCase = (kind, data)=>serializeData[kind](data);
const branchIncludesMorph = (branch, $)=>"morph" in branch ? true : "props" in branch ? Object.values(branch.props).some((prop)=>nodeIncludesMorph((0, _propsTs.propToNode)(prop), $)) : false;
const nodeIncludesMorph = (node, $)=>typeof node === "string" ? $.resolve(node).includesMorph : Object.values($.resolveTypeNode(node)).some((predicate)=>predicate === true ? false : (0, _objectKindsTs.isArray)(predicate) ? predicate.some((branch)=>branchIncludesMorph(branch, $)) : branchIncludesMorph(predicate, $));
